#ifndef __TS_H__
#define __TS_H__

// Variable Type Definitions
#define	UDWORD              unsigned int
#define	DWORD               int
#define	UWORD               unsigned short
#define	WORD                short
#define	UBYTE               unsigned char
#define	BYTE                char

// Constant for codepage conversion
#define CP_JOHAP            1361

// Plugin Version Information
#define	VERSION             "v0.8 "

// System Settings
#define PIC_CLK             1193180.0
#define CLK_DIV_MAX         65535
#define TEMPO_BASE          298295LU

// Player Settings
#define SAMPLING_RATE       44100
#define BUFFER_SZ_MAX       SAMPLING_RATE
#define MAX_BUFFER          20
#define MAX_TRACK           22
#define ROL_MAX_CHANNEL     12

// Override the previous declaration
#ifdef NOERROR
#undef NOERROR
#endif

#define NOERROR             0
#define ERR_MEM             -1
#define ERR_BNOT            -2
#define ERR_FNOT            -3
#define ERR_FEOF            -4

//-------------------------------------------------------

#define SAMPLE_FACTOR       0.995

//-------------------------------------------------------

// IMS command code
#define IMSEVENT_MASK       0xF0
#define IMSCHANNEL_MASK     0x0F
#define IMSEVENT_NOTEON     0x80
#define IMSEVENT_NOTEOFF    0x90
#define IMSEVENT_VOLUME     0xA0
#define IMSEVENT_INSTRUMENT 0xC0
#define IMSEVENT_PITCH      0xE0
#define IMSEVENT_TEMPO      0xF0
#define IMSEVENT_END        0xFC

#define IMSTIMECODE_MAX     0xF8

//-------------------------------------------------------

#define TIMEBASE            240

#define INVALID_PITCH_VALUE 0x7FFF

//-------------------------------------------------------

typedef union
{
  UWORD c;
  UBYTE ch[2];
} WordToByte;

typedef struct
{
	UWORD index;
	BYTE  mark;
	BYTE  name[9];
} BankData;

typedef enum
{
  FILE_SOP = 0,
  FILE_IMS,
  FILE_ROL,
  FILE_NA
} F_TYPE;

typedef enum
{
  INS_4OP = 0,
  INS_2OP,
  INS_BDRUM = 6,
  INS_SDRUM,
  INS_TOMTOM,
  INS_CYMBAL,
  INS_HIHAT,
  INS_UNKNOWN,
  INS_NOSOUND
} INST_TYPE;

typedef struct
{
  UBYTE mode;
  UBYTE insname[28];
  UBYTE insdata[23];
} InstrumentData;

// ISS control data structure
#define ISS_TEXT_LINE   64

typedef struct
{
	UWORD tick_iss;
	UBYTE line;
	UBYTE col;
	UBYTE len;

    // ACTUAL TIMINGS
    UDWORD tick;
    UDWORD msec;
} ScriptData;

typedef struct
{
  UBYTE loaded;
  UWORD timecode_index, timecode_running_index, timecode_max, text_max;
  ScriptData *timecode;
  UBYTE *text;
} Script;

// Internal Data (NEVER MODIFY OUTSIDE THIS PLUGIN!!)
typedef struct
{
  // Version Management
  unsigned long sizeOfStructure;

  // PCM
  int time_remaining;
  DWORD sampling_rate;
  DWORD buffer_index, output_index, output_offset, available_buffers;
  DWORD buffer_size[MAX_BUFFER];
  BYTE *ring_buffer[MAX_BUFFER];

  // COMMON
  F_TYPE file_type;
  WORD percussion_mode;
  BYTE filename[512];
  BYTE song_title[64];

  int delay_time, global_volume;
  int play_end, play_end_backup;
  unsigned tempo;
  long elapsed_samples, elapsed_tick;
  WORD ticks_per_beat;
  DWORD ticks;

  // BANK
  int instrument_number;
  InstrumentData *instrument;

  // ISS
  Script iss;

  // SOP
  UWORD total_channels;
  UDWORD channel_size[MAX_TRACK], event_left[MAX_TRACK], event_left_backup[MAX_TRACK];
  BYTE *sop_data[MAX_TRACK], *sop_play_data[MAX_TRACK], operator_type[MAX_TRACK];

  int cur_vol[MAX_TRACK], ticks_left[MAX_TRACK];
  DWORD max_tick[MAX_TRACK];

  // IMS
  UDWORD byte_size, cur_byte;
  UBYTE *ims_data;
  WORD basic_tempo, idcode_st;

  // ROL
  UWORD tempo_counter, *tempo_data, tempo_channel;
  UWORD *ticks_data[ROL_MAX_CHANNEL];
  UWORD *instrument_list[ROL_MAX_CHANNEL];
  UWORD *volume_data[ROL_MAX_CHANNEL];
  UWORD *pitch_data[ROL_MAX_CHANNEL];
  UWORD *instrument_time[ROL_MAX_CHANNEL];
  UWORD instrument_counter_max[ROL_MAX_CHANNEL];
  UWORD volume_counter_max[ROL_MAX_CHANNEL];
  UWORD pitch_counter_max[ROL_MAX_CHANNEL];
  UWORD ticks_counter[ROL_MAX_CHANNEL];
  UWORD instrument_counter[ROL_MAX_CHANNEL];
  UWORD volume_counter[ROL_MAX_CHANNEL];
  UWORD pitch_counter[ROL_MAX_CHANNEL];
} InternalData;

//-------------------------------------------------------

void SetClkRate(unsigned);
int Calculate_Wavelen(InternalData *, int);

//-------------------------------------------------------

void InitCode(void);
void ExitCode(void);
WORD Load_SOP(BYTE *fname, InternalData *info);
void Unload_SOP(InternalData *info);
int Initialize_SOP(InternalData *info, int real);
void Finalize_SOP(InternalData *info);
void Initialize(InternalData *info, int realize);
int Start_SOP(InternalData *info);
void End_SOP(InternalData *info);
void SaveStatus(void);
void ApplyTemporaryStatus(InternalData *info);
void RestoreStatus(void);
int GetData(InternalData *, BYTE *, int);
int SeekToPos(InternalData *info, int offset);

//-------------------------------------------------------

void Update_Time_ISS(InternalData *, long, long);
long Calc_Time(InternalData *);
void Set_Position(InternalData *, long *, long);
void Unload_IMS_ROL(InternalData *);
void Unload_ISS(InternalData *);
void Initialize_ISS(InternalData *);
int Start_IMS_ROL(InternalData *);
void End_IMS_ROL(InternalData *);
int TimeOut_IMS_ROL(InternalData *);
void Load_ISS(InternalData *, char *);
int Load_IMS_ROL(InternalData *, char *);

// ------------------------------------------------------

void *GetFileName(void *str);
void SetClkRate(unsigned count);
void ConvertCodePage(BYTE *string, int maxlen);

#define ReadFileX(a, b, c, d, e)  ReadFile((a), (b), (c), (LPDWORD)(d), (LPOVERLAPPED)(e))

#endif
